﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.ServiceModel;
using HeuristicLab.Services.Hive.Common;
using HeuristicLab.Services.Hive.Common.DataTransfer;
using HeuristicLab.Services.Hive.Common.ServiceContracts;
using DA = HeuristicLab.Services.Hive.DataAccess;

namespace HeuristicLab.Services.Hive {

  /// <summary>
  /// Implementation of the Hive service (interface <see cref="IHiveService"/>).
  /// We need 'IgnoreExtensionDataObject' Attribute for the slave to work. 
  /// </summary>
  [ServiceBehavior(InstanceContextMode = InstanceContextMode.PerCall, IgnoreExtensionDataObject = true)]
  public class HiveService : IHiveService {
    private DataAccess.IHiveDao dao {
      get { return ServiceLocator.Instance.HiveDao; }
    }
    private IAuthenticationManager authen {
      get { return ServiceLocator.Instance.AuthenticationManager; }
    }
    private IAuthorizationManager author {
      get { return ServiceLocator.Instance.AuthorizationManager; }
    }
    private ILifecycleManager lifecycleManager {
      get { return ServiceLocator.Instance.LifecycleManager; }
    }
    private HeartbeatManager heartbeatManager {
      get { return ServiceLocator.Instance.HeartbeatManager; }
    }

    #region Job Methods
    public Guid AddJob(Job job, JobData jobData, IEnumerable<Guid> resourceIds) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return DA.TransactionManager.UseTransaction(() => {
        job.Id = dao.AddJob(job);
        jobData.JobId = job.Id;
        jobData.LastUpdate = DateTime.Now;
        foreach (Guid slaveGroupId in resourceIds) {
          dao.AssignJobToResource(job.Id, slaveGroupId);
        }
        dao.AddJobData(jobData);
        dao.UpdateJobState(job.Id, JobState.Waiting, null, author.UserId, null);
        return jobData.JobId;
      }, false, true);
    }

    public Guid AddChildJob(Guid parentJobId, Job job, JobData jobData) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return DA.TransactionManager.UseTransaction(() => {
        job.ParentJobId = parentJobId;
        return AddJob(job, jobData, dao.GetAssignedResources(parentJobId).Select(x => x.Id));
      }, false, true);
    }

    public Job GetJob(Guid jobId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      return dao.GetJob(jobId);
    }

    public IEnumerable<Job> GetJobs() {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return dao.GetJobs(x => true);
    }

    public IEnumerable<LightweightJob> GetLightweightJobs(IEnumerable<Guid> jobIds) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return dao.GetJobs(x => jobIds.Contains(x.JobId)).Select(x => new LightweightJob(x)).ToArray();
    }

    public IEnumerable<LightweightJob> GetLightweightChildJobs(Guid? parentJobId, bool recursive, bool includeParent) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return GetChildJobs(parentJobId, recursive, includeParent).Select(x => new LightweightJob(x)).ToArray();
    }

    public IEnumerable<LightweightJob> GetLightweightExperimentJobs(Guid experimentId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return dao.GetJobs(x => x.HiveExperimentId == experimentId).Select(x => new LightweightJob(x)).ToArray();
    }

    public JobData GetJobData(Guid jobId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      return dao.GetJobData(jobId);
    }

    public void UpdateJob(Job job) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      DA.TransactionManager.UseTransaction(() => {
        dao.UpdateJob(job);
      });
    }

    public void UpdateJobData(Job job, JobData jobData) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      //DA.TransactionManager.UseTransaction(() => { // cneumuel: try without transaction
      jobData.LastUpdate = DateTime.Now;
      dao.UpdateJob(job);
      dao.UpdateJobData(jobData);
      //}, false, true);
    }

    public void DeleteJob(Guid jobId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      DA.TransactionManager.UseTransaction(() => {
        dao.DeleteJob(jobId);
      });
    }

    public void DeleteChildJobs(Guid parentJobId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      DA.TransactionManager.UseTransaction(() => {
        var jobs = GetChildJobs(parentJobId, true, false);
        foreach (var job in jobs) {
          dao.DeleteJob(job.Id);
          dao.DeleteJobData(job.Id);
        };
      });
    }

    public Job UpdateJobState(Guid jobId, JobState jobState, Guid? slaveId, Guid? userId, string exception) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      return DA.TransactionManager.UseTransaction(() => {
        Job job = dao.UpdateJobState(jobId, jobState, slaveId, userId, exception);

        if (job.Command.HasValue && job.Command.Value == Command.Pause && job.State == JobState.Paused) {
          job.Command = null;
        } else if (job.Command.HasValue && job.Command.Value == Command.Abort && job.State == JobState.Aborted) {
          job.Command = null;
        } else if (job.Command.HasValue && job.Command.Value == Command.Stop && job.State == JobState.Aborted) {
          job.Command = null;
        } else if (jobState == JobState.Paused && !job.Command.HasValue) {
          // slave paused and uploaded the job (no user-command) -> set waiting.
          job = dao.UpdateJobState(jobId, JobState.Waiting, slaveId, userId, exception);
        }

        dao.UpdateJob(job);
        return job;
      });
    }
    #endregion

    #region Job Control Methods
    public void StopJob(Guid jobId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      DA.TransactionManager.UseTransaction(() => {
        var job = dao.GetJob(jobId);
        if (job.State == JobState.Calculating || job.State == JobState.Transferring) {
          job.Command = Command.Stop;
          dao.UpdateJob(job);
        } else {
          if (job.State != JobState.Aborted && job.State != JobState.Finished && job.State != JobState.Failed) {
            job = UpdateJobState(jobId, JobState.Aborted, null, null, string.Empty);
          }
        }
      });
    }

    public void PauseJob(Guid jobId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      DA.TransactionManager.UseTransaction(() => {
        var job = dao.GetJob(jobId);
        if (job.State == JobState.Calculating || job.State == JobState.Transferring) {
          job.Command = Command.Pause;
          dao.UpdateJob(job);
        } else {
          job = UpdateJobState(jobId, JobState.Paused, null, null, string.Empty);
        }
      });
    }

    public void RestartJob(Guid jobId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      DA.TransactionManager.UseTransaction(() => {
        Job job = dao.UpdateJobState(jobId, JobState.Waiting, null, author.UserId, string.Empty);
        job.Command = null;
        dao.UpdateJob(job);
      });
    }
    #endregion

    #region HiveExperiment Methods
    public HiveExperiment GetHiveExperiment(Guid id) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return dao.GetHiveExperiments(x =>
        x.HiveExperimentId == id
        && (x.OwnerUserId == author.UserId || x.HiveExperimentPermissions.Count(hep => hep.Permission != Permission.NotAllowed && hep.GrantedUserId == author.UserId) > 0)
      ).FirstOrDefault();
    }

    public IEnumerable<HiveExperiment> GetHiveExperiments() {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return dao.GetHiveExperiments(x => x.OwnerUserId == author.UserId || x.HiveExperimentPermissions.Count(hep => hep.Permission != Permission.NotAllowed && hep.GrantedUserId == author.UserId) > 0);
    }

    public IEnumerable<HiveExperiment> GetAllHiveExperiments() {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator);
      return dao.GetHiveExperiments(x => true);
    }

    public Guid AddHiveExperiment(HiveExperiment hiveExperimentDto) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return DA.TransactionManager.UseTransaction(() => {
        hiveExperimentDto.OwnerUserId = author.UserId;
        hiveExperimentDto.DateCreated = DateTime.Now;
        return dao.AddHiveExperiment(hiveExperimentDto);
      });
    }

    public void UpdateHiveExperiment(HiveExperiment hiveExperimentDto) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      DA.TransactionManager.UseTransaction(() => {
        dao.UpdateHiveExperiment(hiveExperimentDto);
      });
    }

    public void DeleteHiveExperiment(Guid hiveExperimentId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      DA.TransactionManager.UseTransaction(() => {
        HiveExperiment he = dao.GetHiveExperiment(hiveExperimentId);
        dao.DeleteHiveExperiment(hiveExperimentId); // child jobs will be deleted by db-trigger
      });
    }
    #endregion

    #region Login Methods
    public void Hello(Slave slaveInfo) {
      authen.AuthenticateForAnyRole(HiveRoles.Slave);
      DA.TransactionManager.UseTransaction(() => {
        var slave = dao.GetSlave(slaveInfo.Id);

        if (slave == null) {
          dao.AddSlave(slaveInfo);
        } else {
          var dbSlave = dao.GetSlave(slaveInfo.Id);

          dbSlave.Name = slaveInfo.Name;
          dbSlave.Description = slaveInfo.Description;

          dbSlave.Cores = slaveInfo.Cores;
          dbSlave.CpuArchitecture = slaveInfo.CpuArchitecture;
          dbSlave.CpuSpeed = slaveInfo.CpuSpeed;
          dbSlave.FreeCores = slaveInfo.FreeCores;
          dbSlave.FreeMemory = slaveInfo.FreeMemory;
          dbSlave.Memory = slaveInfo.Memory;
          dbSlave.OperatingSystem = slaveInfo.OperatingSystem;

          dbSlave.LastHeartbeat = DateTime.Now;
          dbSlave.SlaveState = SlaveState.Idle;

          // don't update those properties: dbSlave.IsAllowedToCalculate, dbSlave.ParentResourceId 

          dao.UpdateSlave(dbSlave);
        }
      });
    }

    public void GoodBye(Guid slaveId) {
      authen.AuthenticateForAnyRole(HiveRoles.Slave);
      DA.TransactionManager.UseTransaction(() => {
        var slave = dao.GetSlave(slaveId);
        if (slave != null) {
          slave.SlaveState = SlaveState.Offline;
          dao.UpdateSlave(slave);
        }
      });
    }
    #endregion

    #region Heartbeat Methods
    public List<MessageContainer> Heartbeat(Heartbeat heartbeat) {
      authen.AuthenticateForAnyRole(HiveRoles.Slave);
      TriggerLifecycle(false);
      return DA.TransactionManager.UseTransaction(() => heartbeatManager.ProcessHeartbeat(heartbeat));
    }
    #endregion

    #region Plugin Methods
    public Guid AddPlugin(Plugin plugin, List<PluginData> pluginDatas) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return DA.TransactionManager.UseTransaction(() => {
        plugin.UserId = author.UserId;
        plugin.DateCreated = DateTime.Now;

        var existing = dao.GetPlugins(x => x.Hash != null).Where(x => x.Hash.SequenceEqual(plugin.Hash));
        if (existing.Count() > 0) {
          // a plugin already exists.
          throw new FaultException<PluginAlreadyExistsFault>(new PluginAlreadyExistsFault(existing.Single().Id));
        }

        Guid pluginId = dao.AddPlugin(plugin);
        foreach (PluginData pluginData in pluginDatas) {
          pluginData.PluginId = pluginId;
          dao.AddPluginData(pluginData);
        }
        return pluginId;
      });
    }

    public Plugin GetPlugin(Guid pluginId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      return dao.GetPlugin(pluginId);
    }

    public IEnumerable<Plugin> GetPlugins() {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      return dao.GetPlugins(x => x.Hash != null);
    }

    public IEnumerable<PluginData> GetPluginDatas(List<Guid> pluginIds) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client, HiveRoles.Slave);
      List<PluginData> pluginDatas = new List<PluginData>();

      return DA.TransactionManager.UseTransaction(() => {
        foreach (Guid guid in pluginIds) {
          pluginDatas.AddRange(dao.GetPluginDatas(x => x.PluginId == guid).ToList());
        }
        return pluginDatas;
      });
    }

    #endregion

    #region Slave Methods
    public Guid AddSlave(Slave slave) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return DA.TransactionManager.UseTransaction(() => dao.AddSlave(slave));
    }

    public Guid AddSlaveGroup(SlaveGroup slaveGroup) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return DA.TransactionManager.UseTransaction(() => dao.AddSlaveGroup(slaveGroup));
    }

    public Slave GetSlave(Guid slaveId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return dao.GetSlave(slaveId);
    }

    public SlaveGroup GetSlaveGroup(Guid slaveGroupId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return dao.GetSlaveGroup(slaveGroupId);
    }

    public IEnumerable<Slave> GetSlaves() {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return dao.GetSlaves(x => true);
    }

    public IEnumerable<SlaveGroup> GetSlaveGroups() {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return dao.GetSlaveGroups(x => true);
    }

    public void UpdateSlave(Slave slave) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      DA.TransactionManager.UseTransaction(() => {
        dao.UpdateSlave(slave);
      });
    }

    public void UpdateSlaveGroup(SlaveGroup slaveGroup) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      DA.TransactionManager.UseTransaction(() => {
        dao.UpdateSlaveGroup(slaveGroup);
      });
    }

    public void DeleteSlave(Guid slaveId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      DA.TransactionManager.UseTransaction(() => {
        dao.DeleteSlave(slaveId);
      });
    }

    public void DeleteSlaveGroup(Guid slaveGroupId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      DA.TransactionManager.UseTransaction(() => {
        dao.DeleteSlaveGroup(slaveGroupId);
      });
    }

    public void AddResourceToGroup(Guid slaveGroupId, Guid resourceId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      DA.TransactionManager.UseTransaction(() => {
        var resource = dao.GetResource(resourceId);
        resource.ParentResourceId = slaveGroupId;
        dao.UpdateResource(resource);
      });
    }

    public void RemoveResourceFromGroup(Guid slaveGroupId, Guid resourceId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      DA.TransactionManager.UseTransaction(() => {
        var resource = dao.GetResource(resourceId);
        resource.ParentResourceId = null;
        dao.UpdateResource(resource);
      });
    }

    public Guid GetResourceId(string resourceName) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator, HiveRoles.Client);
      return DA.TransactionManager.UseTransaction(() => {
        var resource = dao.GetResources(x => x.Name == resourceName).FirstOrDefault();
        if (resource != null) {
          return resource.Id;
        } else {
          return Guid.Empty;
        }
      });
    }

    public void TriggerLifecycle(bool force) {
      // use a serializable transaction here to ensure not two threads execute this simultaniously (mutex-lock would not work since IIS may use multiple AppDomains)
      DA.TransactionManager.UseTransaction(() => {
        DateTime lastCleanup = dao.GetLastCleanup();
        if (force || DateTime.Now - lastCleanup > TimeSpan.FromSeconds(59)) {
          dao.SetLastCleanup(DateTime.Now);
          lifecycleManager.Cleanup();
        }
      }, true);
    }
    #endregion

    #region Helper Methods
    private IEnumerable<Job> GetChildJobs(Guid? parentJobId, bool recursive, bool includeParent) {
      var jobs = new List<Job>(dao.GetJobs(x => parentJobId == null ? !x.ParentJobId.HasValue : x.ParentJobId.Value == parentJobId));

      if (recursive) {
        var childs = new List<Job>();
        foreach (var job in jobs) {
          childs.AddRange(GetChildJobs(job.Id, recursive, false));
        }
        jobs.AddRange(childs);
      }

      if (includeParent) jobs.Add(GetJob(parentJobId.Value));
      return jobs;
    }
    #endregion

    #region Appointment Methods
    public Guid AddAppointment(Appointment appointment) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator);
      return DA.TransactionManager.UseTransaction(() => dao.AddAppointment(appointment));
    }

    public void DeleteAppointment(Guid appointmentId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator);
      DA.TransactionManager.UseTransaction(() => {
        dao.DeleteAppointment(appointmentId);
      });
    }

    public void UpdateAppointment(Appointment appointment) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator);
      DA.TransactionManager.UseTransaction(() => {
        dao.UpdateAppointment(appointment);
      });
    }

    public IEnumerable<Appointment> GetScheduleForResource(Guid resourceId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator);
      return DA.TransactionManager.UseTransaction(() => dao.GetAppointments(x => x.ResourceId == resourceId));
    }

    public IEnumerable<Job> GetJobsByResourceId(Guid resourceId) {
      authen.AuthenticateForAnyRole(HiveRoles.Administrator);
      return DA.TransactionManager.UseTransaction(() => dao.GetJobsByResourceId(resourceId));
    }
    #endregion
  }
}
